/*
 * aimage.cpp:
 *
 * Image a physical drive connected through an IDE, SCSI, USB or Firewire interface
 * and write the results to either a raw or AFF file.
 *
 * Automatically calculate the MD5 & SHA-1 while image is made.
 * Handle read errors in an intelligent fashion. 
 * Display the results. 
 */

#include "afflib.h"			
#include "afflib_i.h"			// grab the internal versions
#include "aimage.h"
#include "timer.h"
#include "imager.h"

#ifdef __FreeBSD__
#include <sys/param.h>
#include <sys/mount.h>
#define HAS_STATFS
#endif

char  cmd_attach[255]={0};		// to detach a device that was attached
char  cmd_detach[255]={0};		// to detach a device that was attached
bool  finishing = false;
int   compressing=0;
timer write_time;
timer total_time;
int  compression_level = AF_COMPRESSION_DEFAULT;// default compression level
int  compression_type  = AF_COMPRESSION_ZLIB;	// 
int  scsi_bus = -1;
int  scsi_tid = -1;
int  scsi_lun = -1;
int  scsi_pass = -1;
int  ata_dev  = -1;
char *opt_title = "IMAGING";
bool imaging_failed = true;
char *command_line = 0;			// what is typed

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology Corp.
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */



char *progname = "aimage";

int  opt_datasegsize = 16*1024*1024;	// default seg size --- 16MB
int  opt_append = 0;
int  opt_quiet = 0;
int  opt_silent = 0;
int  opt_beeps  = 1;		// beep when done
int  opt_readsectors = 512;	// read in 256K chunks
int  opt_zap = 0;
char *opt_config=0;
int  opt_error = 0;
int  opt_retry = 5;
int  opt_reverse = 0;
int  opt_fast_quit = 0;
char *infile = 0;			// data source
time_t  start_time = 0;
const char *logfile_fname = 0;
FILE *logfile;

int opt_skip = 0;			// what to skip on input
int opt_skip_sectors = 0;		// are we skipping sectors?
int opt_compress = 1;
int opt_preview = 1;			// give me a little show

char device_model[256] = {0};
char serial_number[256] = {0};
char firmware_revision[256] = {0};


void usage()
{
    putchar('\n');
    printf("usage: %s [options] INPUT [OUTFILE] --- image indev to outfile\n",progname);
    printf("INPUT may be any of these:\n");
    printf("  A device (e.g. /dev/disk1)\n");
    printf("  - (or /dev/stdin, for standard input)\n");
    printf("  listen:nnnn       Listen on TCP port nnnn\n");

    printf("OUTFILE may be:\n");
    printf("  outfile.aff --- image to the AFF file outfile\n");

    bold("\nGeneral Options:\n");
    printf("  --quiet, -q           -- No interactive statistics.\n");
    printf("  --silent, -Q          -- No output at all except for errors.\n");
    printf("  --readsectors=nn, -R nnnn,   -- set number of sectors to read at once (default %d)\n",
	                                      opt_readsectors);
    printf("  --skip=nn[s], -k nn   -- Skip nn bytes [or nns for sectors] in input file\n");
    printf("  --no_beeps, -B        -- Don't beep when imaging is finished.\n");
    printf("  --logfile=fn          -- Where to write a log. Specify /dev/null to suppress log.\n");
    printf("                           default is %s.\n",logfile_fname);
    printf("  --info, -I            -- print configuration info.\n");
    printf("  --preview, -p         -- view some of the data as it goes by.\n");
    printf("  --no_preview, -P      -- do not show the preview.\n");

    bold("\nExisting File Options:\n");
    printf("  --append, -a          -- Append to existing file\n");
    printf("  --zap, -z             -- Erase outfile(s) before writing\n");

    bold("\nRaw Output Options:\n");
    printf("  --raw=fname, -r fname -- write block-by-block output to fname.\n");

    bold("\nAFF Options:\n");
    printf("  --outfile=fname, -ofname  -- write an AFF file.\n");
    printf("  --image_segsize=nnn, -S nnnn   -- set the AFF image_segsize (default %d)\n",
	   opt_datasegsize);
    printf("                     otherwise takes off where it left off.\n");
    printf("  --config=fn,  -c file -- Read question config from file (default %s)\n",opt_config);
    printf("  --make_config=fn      -- write a sample config file\n");
    printf("  --no_compress, -x     -- Do not compress. Useful on slow machines.\n");
    printf("  --compression=n, -Xn  -- Set the compression level\n");

    bold("\nError Recovery Options:\n");
    printf("  --error=0, -e0  -- Standard error recovery:\n");
    printf("                     Read disk 256K at a time until there are 5 errors in a row.\n");
    printf("                     Then go to the end of the disk and read backwards\n");
    printf("                     until there are 5 erros in a row. Then stop.\n");
    printf("  --error=1  -e1  -- Stop reading at first error.\n");
    printf("  --retry=nn -tnn -- change retry count from 5 to nn\n");
    printf("  --reverse, -v   -- Scan in reverse from the beginning.\n");

    bold("\nOther:\n");
    printf("  --free dev, -F dev  -- Just print the free space on DEV and exit\n");
    printf("  --help, -h      -- Print this message.\n");
    printf("  --fast_quit, -Z -- Make ^c just exit immediately.\n");
    printf("  --allow_regular,   -E -- allow the imaging of a regular file\n");
    printf("  --title=s, -T s -- change title to s (from IMAGING)\n");
    exit(0);
}


static struct option longopts[] = {
    { "raw",	       required_argument,  NULL, 'r'}, // can use this for flags, but not for longs
    { "outfile",       required_argument,  NULL, 'o'},
    { "quiet",         no_argument,        &opt_quiet, 'q'},
    { "silent",        no_argument,        NULL, 'Q'},
    { "readsectors",   required_argument,  NULL, 'R'},
    { "image_segsize", required_argument,  NULL, 'S'},
    { "zap",           no_argument,        &opt_zap, 'z'},
    { "config",        required_argument,  NULL, 'c'},
    { "help",          no_argument,        NULL, 'h'},
    { "retry",         required_argument,  NULL, 't'},
    { "no_compress",   no_argument,        NULL, 'x'},
    { "compression",   required_argument,  NULL, 'X'},
    { "no_beeps",      no_argument,        NULL, 'B'},
    { "append",        no_argument,        NULL, 'a'},
    { "make_config",   required_argument,  NULL, 'm'},
    { "logfile",       required_argument,  NULL, 'l'},
    { "info",          no_argument,        NULL, 'I'},
    { "reverse",       no_argument,        NULL, 'v'},
    { "fast_quit",     no_argument,        NULL, 'Z'},
    { "allow_regular", no_argument,        NULL, 'E'},
    { "title",         required_argument,  NULL, 'T'},
    { "free",          required_argument,  NULL, 'F'},
    { "preview",       no_argument,        NULL, 'p'},
    { "no_preview",    no_argument,        NULL, 'P'},
    {0,0,0,0}
};


/****************************************************************
 *** isleep(): An informative sleep...
 ****************************************************************/
void isleep(int s)
{
  printf("isleep %d\n",s);
  for(int i=0;i<s;i++){
    printf("\rSleeping for %d seconds; %d left...",s,s-i);
    fflush(stdout);
    sleep(1);
  }
  printf("\r%50s\r","");
  fflush(stdout);
}


imager *im;
uint64 total_bytes_compressed = 0;
uint64 total_bytes_uncompressed = 0;


void print_freespace(int fd)
{
#ifdef HAS_STATFS
    struct statfs sbuf;
    if(fstatfs(fd,&sbuf)==0){
	char buf[1024];
	sprintf(buf,"Free space on %s:",sbuf.f_mntfromname);

	printf("%25s ",buf);
	float percent_free = sbuf.f_bavail * 100.0 / sbuf.f_blocks;
	uint64 free_mbytes = (unsigned long long)sbuf.f_bavail * sbuf.f_bsize / (1024 * 1024);
	comma_printf(free_mbytes,0);
	printf(" MB (%5.2f%%)", percent_free);
    }
#endif
}

void print_freespace(char *fname)
{
    int fd = open(fname,O_RDONLY,0666);
    if(fd<0) err(1,fname);
    print_freespace(fd);
}

void compression_callback(int segnum,int bytes_uncompressed,int bytes_compressed)
{
    if(bytes_uncompressed==0 && bytes_compressed==0){
	compressing = true;
	write_time.start();
    }
    else{
	compressing = false;
	write_time.stop();
    }
	

    total_bytes_compressed  += bytes_compressed;
    total_bytes_uncompressed  += bytes_uncompressed;
    if(opt_quiet==0 && opt_silent==0 && finishing==false){
	my_refresh();
    }
}



/* sig_intr:
 * Close the AFF file if possible
 * If this is the second time we were called, just exit...
 */

int depth = 0;
void sig_intr(int arg)
{
    int rows = tgetnum("li");

    if(++depth==1){
      // don't get confused...
	af_enable_compression(im->af, compression_type,compression_level,0);
	gotorc(rows-4,0);
	printf("Interrupt!\n");
	printf("\n\n\n");
	if(opt_fast_quit){
	    printf("*** FAST QUIT ***\n");
	    exit(1);
	}
	printf("Closing output file...\n");
	if(im->af){
	    if(af_close(im->af)){
		warn("Can't close file '%s'\n",af_filename(im->af));
	    }
	}
    }
    fflush(stdout);
    exit(1);
}

void sig_cont(int arg)
{
    beep();beep();
    screen_cleared = false;		// redraw the screen
}


char lastchar(const char *str)
{
    return str[strlen(str)-1];
}


/* Special logic for opening a FreeBSD device... */
void Sleep(int msec)
{
    usleep(msec*1000);
}



void make_config(char *fname)
{
    if(access(fname,F_OK)==0){
	fprintf(stderr,"make_config: %s exists.\n",fname);
	exit(1);
    }
    FILE *f = fopen(fname,"w");
    if(!f) err(1,"make_config(%s) ",fname);
    fputs("#\n",f);
    fputs("#\n",f);
    fputs("#\n",f);
    fputs("# Sample config file for aimage...\n",f);
    fputs("# Syntax:  ask <segname> <question>\n",f);
    fputs("#\n",f);
    fputs("ask name Your Name:\n",f);
    fputs("ask date_acquired Date this drive was acquired (yyyy-mm-dd):\n",f);
    fputs("ask case_number Case Number:\n",f);
    fputs("#\n",f);
    fputs("# Feel free to add your own!\n",f);
    fclose(f);
    printf("%s created\n",fname);
    exit(0);
}


void process_config(AFFILE *af)
{
    FILE *f = fopen(opt_config,"r");	// open the config file to find questions to ask
    if(!f) return;			// get back to the user
    while(!feof(f)){
	char buf[1024];			// a righteous buffer for questions and answers
	char *sep = " \t";
	char *last;
	memset(buf,0,sizeof(buf));
	fgets(buf,sizeof(buf)-1,f);
	
	char *cc = index(buf,'\n');
	if(cc) *cc = '\000';		// remove the \n

	char *cmd = strtok_r(buf,sep,&last);

	if(!cmd || strcmp(cmd,"ask")!=0) continue; // not an ask command
	char *segname = strtok_r(NULL,sep,&last);
	if(!segname){
	    fprintf(stderr,"error in config file. No segnament name in: %s\n",buf);
	    continue;
	}
	
	/* Ask the question and get the response */
	char *val = readline(last);

	/* And write the response into the segment */
	af_write_seg(af,segname,0,val,strlen(val));
	free(val);
    }
    fclose(f);
}




/*
 * open_dev(char outdev[MAXPATHLEN],char *indev)
 * Try to open the friendly device name.
 * If successful, return the actual device in outdev and the FD.
 * 
 * If the device can be detected but not mounted (common with some
 * broken IDE drives), return fd==65536 (FD_IDENT). 
 * This says that we can't open it, but should ident it.
 */

#define FD_IDENT 65536

int  open_dev(char outdev[MAXPATHLEN],char *indev)
{
    /****************************************************************
     *** Check for ata%d or ide%d
     ****************************************************************/

    outdev[0] = 0;			// for starters
    ata_dev = -1;
    sscanf(indev,"ata%d",&ata_dev);	// try to find ata0
    if(ata_dev==-1){
	sscanf(indev,"ide%d",&ata_dev);	// try to find ide0
    }

    if(ata_dev != -1){			// if we found the device
	/* Create the attach and detach commands */

	char dev0[64];			// space for the first channel
	char dev1[64];			// space for the second channel
	char *dev[2] = {dev0,dev1};

	make_ata_attach_commands(cmd_attach,cmd_detach,dev0,dev1,ata_dev);
	system(cmd_detach);	// make sure we are detached first

	int i;
	for(i=0;i<10;i++){
	    int delay = i*3;
	    printf("\nOpening special ATA Bus #%d...\n",ata_dev);
	    if(i>0){
		printf("Attempt %d out of %d.\n",i+1,10);
	    }
	    printf("# %s\n",cmd_attach);
	    system(cmd_attach);
	    
	    /* See if we found the device */
	    for(int d=0;d<2;d++){
		if(access(dev[d],F_OK)==0){
		    if(access(dev[d],R_OK)){
			// don't have permission to read it.
			// this is bad
			err(1,outdev);
		    }
		    if(delay){
			printf("Waiting %d second%s for %s to spin up...\n",
			       delay,delay==1?"":"s",outdev);  
			isleep(delay);
		    }	
		    strcpy(outdev,dev[d]); // we will try this one
		    int fd = open(outdev,O_RDONLY);
		    if(fd>0){
			/* The device was successfully opened. */
			return fd;		// got it!
		    }
		    perror(outdev);		// give a warning
		}
	    }
	    printf("Detaching device and trying again...\n");
	    printf("# %s\n",cmd_detach);
	    system(cmd_detach);
	    isleep(delay);
	}
	/* Been through too many times. Did we get a device?
	 * If so, just ident it...
	 */
	if(outdev[0]){
	    return FD_IDENT;
	}
    }

    /****************************************************************
     *** Check for scsi%d
     *** In our testing with FreeBSD, there is no advantage to repeatedly
     *** attempting to attach or detach...
     ****************************************************************/
    if(sscanf(outdev,"scsi%d",&scsi_bus)==1){
	if(scsi_attach(outdev,sizeof(outdev),scsi_bus)==0){
	    int fd = open(outdev,O_RDONLY);
	    if(fd>0){
		return fd;
	    }
	}
    }

    outdev[0] = 0;			// in the end, we found nothing
    return -1;
}


void print_info()
{
    for(int i=0;i<2;i++){
	char master[64];
	char slave[64];
	printf("Device %d:\n",i);
	make_ata_attach_commands(cmd_attach,cmd_detach,master,slave,i);
	printf("  attach with '%s'\n",cmd_attach);
	printf("  deatch with '%s'\n",cmd_detach);
	printf("  master is %s\n",master);
	printf("  slave is %s\n",slave);
	printf("\n");
    }
    exit(0);
}

char *append(char *base,char *str)
{
    base = (char *)realloc(base,strlen(base)+strlen(str)+1);
    strcat(base,str);
    return base;
}

unsigned char final_md5[16];
unsigned char final_sha1[20];


void final_report(imager *im)
{
    /* Let's make the user feel good... */
    gotorc(tgetnum("li"),0);
    printf("\n\n\n");
    bold("****************************************************************");
    putchar('\n');
    printf("FINAL IMAGING REPORT\n");
    printf("\n");
    printf("%20sInput: ",""); bold(infile); putchar('\n');
    if(device_model[0]){
      printf("%13sDevice Model: ","");
      bold(device_model);
      putchar('\n');
    }
    if(serial_number[0]){
      printf("%12sSerial Number: ","");
      bold(serial_number);
      putchar('\n');
    }
    if(im->af){
	printf("%10sAFF Output file: ","");
	bold(af_filename(im->af));
	putchar('\n');
    }
    if(im->fout){
	printf("%12sRaw Output file: ","");
	bold(im->fout_fname);
	putchar('\n');
    }

    putchar('\n');
    
    printf("         Total bytes read: ");
    comma_printf(im->total_bytes_read,15);
    putchar('\n');
    printf(" Compressed bytes written: ");
    comma_printf(total_bytes_compressed,15);
    putchar('\n');
    
    
    char print_buf[256];
    printf("\n");
    if(im->hash_invalid==false){
	printf("raw image md5:  %s\n",
	       af_hexbuf(print_buf,sizeof(print_buf),final_md5,16,1));
	
	printf("raw image sha1: %s\n",
	       af_hexbuf(print_buf,sizeof(print_buf),final_sha1,20,1));
    }

    if(imaging_failed){
      printf("\nTHIS DRIVE COULD NOT BE IMAGED DUE TO A HARDWARE FAILURE.\n");
    }
}

/* Return 1 if the PID is running, 0 if it is not
 * in an operating-system independent kind of way.
 */
int checkpid(int pid)
{
    char buf[1024];
    int  found = 0;
    sprintf(buf,"ps %d",pid);
    FILE *f = popen(buf,"r");
    while(!feof(f)){
	char *cc = buf;
	fgets(buf,sizeof(buf),f);
	buf[sizeof(buf)-1] = 0;
	if(atoi(cc)==pid) found = 1;
    }
    pclose(f);
    return found;
}

/* getlock(char *infile):
 * See if another copy of aimage is imaging this infile...
 * A lock file is created in /tmp that has the PID of the aimage process.
 * Return 0 if okay, -1 if not okay
 */

int getlock(char *infile)
{
    /* If the file exists and the PID in the file is running,
     * can't get the lock.
     */
    char lockfile[MAXPATHLEN];
    sprintf(lockfile,"/tmp/aimge.%s.lock",infile);
    if(access(lockfile,F_OK)==0){
	/* Lockfile exists. Get it's pid */
	char buf[1024];
	FILE *f = fopen(lockfile,"r");
	if(!f){
	    perror(lockfile);		// can't read lockfile...
	    return -1;
	}
	fgets(buf,sizeof(buf),f);
	buf[sizeof(buf)-1] = 0;
	int pid = atoi(buf);
	if(checkpid(pid)==0){
	    /* PID is not running; we can delete the lockfile */
	    if(unlink(lockfile)){
		err(1,"could not delete lockfile %s: ",lockfile);
	    }
	}
	/* PID is running; generate error */
	errx(1,"%s is locked by process %d\n",infile,pid);
    }
    FILE *f = fopen(lockfile,"w");
    if(!f){
	err(1,lockfile);
    }
    fprintf(f,"%d\n",getpid());		// save our PID.
    fclose(f);
    return 0;
}

void open_logfile(const char *lfn)
{
    /* Create a logfile... */
    char buf[1024];
    sprintf(buf,"%s/.aimage_log",getenv("HOME"));
    logfile_fname = strdup(buf);
    logfile = fopen(logfile_fname,"a");
}

int socket_listen(int port)
{
    struct sockaddr_in local;
    struct sockaddr_in remote;
    socklen_t rsize = sizeof(remote);

    int sock = socket(AF_INET,SOCK_STREAM,IPPROTO_IP);    /* Open a listening socket ... */
    memset(&local,0,sizeof(local));
    memset(&remote,0,sizeof(remote));
#ifndef linux				// have I mentioned how much I hate linux?
    local.sin_len = sizeof(sockaddr_in);
#endif
    local.sin_family = AF_INET;
    local.sin_port   = htonl(port);	// listen on port 5555
    if(bind(sock,(sockaddr *)&local,sizeof(local))) err(1,"bind");
    if(listen(sock,0)) err(1,"listen");		// listen, and only accept one
    printf("Listening for connection on port %d...\n",port);
    int s2 = accept(sock,(sockaddr *)&remote,&rsize);
    if(s2<0) err(1,"accept");
    infile = strdup(inet_ntoa(remote.sin_addr));
    printf("Connection accepted from %s\n",infile);
    return s2;
}


/* image_fd:
 * Actually do the imaging...
 */
int image_fd(int ifd,const char *outfile_raw,const char *outfile_aff)
{
    if(im->set_input(ifd)){
	return -1;		// couldn't image?
    }

    /* See if the skipping makes sense */
    if(!opt_skip_sectors){
	if(opt_skip % im->sector_size != 0){
	    fprintf(stderr,
		    "Skipping must be an integral multiple of sector size "
		    "(%d bytes)\n",im->sector_size);
	    return -1;
	}
	opt_skip /= im->sector_size;	// get the actuall offset
    }

    FILE *fout = 0;
    if(outfile_raw){
	fout = fopen(outfile_raw,"wb");
	if(!fout){
	  err(1,"fopen(%s)",outfile_raw);
	}
	im->set_fout(fout);
	strcpy(im->fout_fname,outfile_raw);
    }

    int starting_direction = 1;
    if(opt_reverse) starting_direction = -1;

    /****************************************************************
     *** Start imaging
     ****************************************************************/

    if(opt_beeps) putchar('\007');			// bee
    total_time.start();
    signal(SIGINT,sig_intr);	// set the signal handler
    im->hash_setup();		// get ready...
    im->image_loop(opt_skip,im->total_sectors,starting_direction,
		   opt_readsectors,opt_error); // start the process
    signal(SIGINT,0);		// unset the handler
    total_time.stop();

    signal(SIGCONT,sig_cont);

    /****************************************************************
     *** Finished imaging
     ****************************************************************/

    /* Calculate the final MD5 and SHA1 */
    MD5_Final(final_md5,&im->md5);
    SHA1_Final(final_sha1,&im->sha);

    /* Finish up */
    int rows = tgetnum("li");

    finishing = true;
    gotorc(rows-2,0);
    tputs(clr_eos,0,putchar);
    printf("\n\n");

    if(fout){
	if(fclose(fout)){
	    warn("fclose(%s)",outfile_raw);
	    fprintf(stderr,
		    "This shouldn't happen. '%s' may be corrupt.\n",
		    outfile_raw);
	}
	fout=0;
    }
    return 0;
}


void aimage(int ifd,const char *outfile_raw,const char *outfile_aff)
{
    if(opt_zap){
	if(outfile_raw) unlink(outfile_raw);	// if we are zapping
	if(outfile_aff) unlink(outfile_aff);
    }

    /* If either file exists and we are not appending, give an error */
    if(!opt_append){
	if(access(outfile_raw,F_OK)==0) errx(1,"%s: file exists",outfile_raw);
	if(access(outfile_aff,F_OK)==0) errx(1,"%s: file exists",outfile_aff);
    }

    /* If the user is imaging to an AFF file,
     * open it and try to ident the drive.
     * Drive ident is not done if writing to a raw file, because
     * there is no place to store the ident information. This will be changed
     * when we can write an XML log.
     */
    AFFILE *af=0;
    if(outfile_aff){
	if(af_identify(outfile_aff)==0){
	    fprintf(stderr,"%s exists and is not an AFF file.\n",outfile_aff);
	    fprintf(stderr,"Delete it or move it first.\n");
	    exit(-1);
	}

	af = af_open(outfile_aff,O_CREAT|O_RDWR,0666);
	if(!af){
	    err(1,"af_open %s: ",outfile_aff);
	}
	af->logfile = logfile;
	/* If the segment size hasn't been set, then set it */
	af_set_dataseg_sizes(af,opt_datasegsize,(uint64)0);		// 
	af_enable_writing(af,1);	// we will be writing this file
	af_enable_compression(af,
			      compression_type,compression_level,
			      compression_callback);
	process_config(af);		// process the config file for questions
	ident(infile,af);		// ident the drive
	af_update_seg(af,AF_IMAGING_COMMAND_LINE,0,command_line,strlen(command_line),1);
	af_update_seg(af,AF_IMAGING_DEVICE,0,infile,strlen(infile),1);
	unsigned char bit128[16];
	RAND_pseudo_bytes(bit128,sizeof(bit128));
	af_update_seg(af,AF_IMAGE_GID,0,bit128,sizeof(bit128),1);

	time_t t = time(0);
	char buf[64];
	strftime(buf,sizeof(buf),"%Y-%m-%d %H:%M:%S",localtime(&t));
	af_update_seg(af,AF_IMAGING_DATE,0,buf,strlen(buf),1);

	/* TK: date_acquired, accession_gid */
    }

    im->retry_count = opt_retry;

    if(opt_quiet ==0 && opt_silent==0){
	im->status = my_refresh;
    }

    if(af) im->set_af(af);

    /* Here is where the imaging takes place.
     * Do it unless ifd==FD_IDENT, which is the fictitious FD.
     */


    nodelay(stdscr,1);
    cbreak();
    noecho();
    if(ifd != FD_IDENT){
      if(image_fd(ifd,outfile_raw,outfile_aff)==0){
	imaging_failed = false;
      }
    }
    echo();
    nocbreak();


    /* AFF Cleanup... */
    if(af){
	if(im->hash_invalid==false){
	    if(af_update_seg(af,AF_MD5,0,final_md5,16,0)){
		perror("Could not update AF_MD5");
	    }
	    if(af_update_seg(af,AF_SHA1,0,final_sha1,20,0)){
		perror("Could not update AF_SHA1");
	    }
	}
	else {
	    af_del_seg(af,AF_MD5);	// because it is not valid
	    af_del_seg(af,AF_SHA1);
	}
	if(af_update_segq(af,AF_BADSECTORS, im->bad_sectors_read,1)){
	    perror("Could not update AF_BADSECTORS");
	}
	if(af_update_segq(af,AF_BLANKSECTORS, im->total_blank_sectors,1)){
	    perror("Could not update AF_BLANKSECTORS");
	}

	if(af_close(af)){
	    perror("af_close");
	    fprintf(stderr,"af_close failed. This shouldn't happen.\n");
	    fprintf(stderr,"Run 'ainfo -v %s' to see if %s is corrupt.\n",
		    outfile_aff,outfile_aff);
	}
	af = 0;
    }

}


int main(int argc,char **argv)
{
    /* build the option string */
    const char *outfile_raw =0 ;
    const char *outfile_aff =0;
    char optstring[256];
    optstring[0] = 0;
    int i;
    char *cc = optstring;
    for(i=0;longopts[i].name;i++){
	switch(longopts[i].has_arg){
	case no_argument:
	    *cc++ = longopts[i].val;
	    break;
	case required_argument:
	case optional_argument:
	    *cc++ = longopts[i].val;
	    *cc++ =':';
	    break;
	default:
	    printf("unknown value for longopts[%d].has_arg=%d\n",
		   i,longopts[i].has_arg);
	    exit(9);
	}
    }
    *cc++ = 0;

    /* Figure out the command line that was used to run this program */
    command_line = (char *)malloc(1);
    command_line[0] = 0;
    for(int i=0;i<argc;i++){
	if(i>0) command_line = append(command_line," ");
	command_line = append(command_line,argv[i]);
    }

    /* Create the imager and get the program name... */
    im = new imager();

    initscr();    /* Set up curses */
    setupterm((char *)0,1,(int *)0);	// set up termcap; is this necessary?
    progname = argv[0];


    for(int i=0;i<1000;i++){
	for(int j=0;j<24;j++){
	    char buf[1024];
	    sprintf(buf,"i=%d j=%d  just another test",i,j);
	    mvaddstr(j,j,buf);
	    refresh();
	}
    }

    int ch;
    while ((ch = getopt_long_only(argc, argv, optstring,longopts,NULL)) != -1) {
	switch (ch) {
	case 'r': outfile_raw = optarg;	    break;
	case 'o': outfile_aff = optarg;	    break;

	case 'q': opt_quiet++;		break; 
	case 'Q': opt_silent++;		break;
	case 'R': opt_readsectors = atoi(optarg); break;
	case 'z': opt_zap ++;		    break;
	case 'a': opt_append ++;	    break;
	case 'k':
	    opt_skip  = atoi(optarg);
	    if(lastchar(optarg)=='s'){
		opt_skip_sectors = 1;
	    }
	    break;

	case 'S': opt_datasegsize = atoi(optarg); break;

	case 'c': opt_config = optarg; break;

	case 'x':
	    opt_compress = 0;
	    compression_level = AF_COMPRESSION_NONE;
	    compression_type  = AF_COMPRESSION_NONE;
	    break;

	case 'X': compression_level = atoi(optarg); break;

	case 'e': opt_error = atoi(optarg); break;

	case 't': opt_retry = atoi(optarg); break;

	case 'B': opt_beeps = 0; break;

	case 'm': make_config(optarg); break;
	    
	case 'I': print_info(); break;

	case 'v': opt_reverse = 1; break;

	case 'Z': opt_fast_quit = 1; break;

	case 'E': im->allow_regular=1; break;

	case 'T':
	    opt_title = optarg;
	    break;

	case 'F': print_freespace(optarg); exit(1);

	case 'p': opt_preview = 1;break;
	case 'P': opt_preview = 0;break;
	    
	case 'h':
	case '?':
	default:
	    usage();
	case 0:
	    break;
	}
    }

    argc -= optind;
    argv += optind;

    if(argc<1){
	usage();
	exit(0);
    }

    open_logfile(logfile_fname);

    infile = argv[0];

    argc--;
    argv++;

    /* IF there is an additional output file,
     * and outfile_aff has not been specified, use that
     */
    if(outfile_aff==0 && argc>0){
	outfile_aff = argv[0];
	argc--;
	argv++;
    }

    start_time = time(0);		// excellent!

    /* First, open the input file.
     * If the file specified by the user works, just open it.
     * If the file doesn't work, see if it is a operating system
     * specific filename such as "ide0" or "ata0", in which the
     * operating system-specific code will attempt to attach the
     * device.
     */
    int ifd = -1;
    int port;
    char outdev[MAXPATHLEN];

    if(strcmp(infile,"-")==0){
	ifd = 0;			// file descriptor 0 is stdin
    }
    if(sscanf(infile,"listen:%d",&port)==1){
	ifd = socket_listen(port);
    }

    if(ifd<0){				// if hasn't been set, try to open the file
	ifd = open(infile,O_RDONLY);
    }
    if(ifd<0){				
	/* Attempt to open infile failed; check for a special
	 * device name...
	 */
	ifd = open_dev(outdev,infile);
	if(ifd>0) infile = outdev;	// switch over
    }

    /* If we haven't been able to open the infile, give up. */
    if(ifd<0){
	perror(infile);
	exit(1);
    }

    aimage(ifd,outfile_raw,outfile_aff);
    final_report(im);

#ifdef __FreeBSD__
    /* Need to detach if we are using freebsd and attached */
    if(cmd_detach[0]){
	printf("Detaching drive...\n");
	system(cmd_detach);
    }
#endif    
    printf("\n");
    if(opt_beeps && !opt_quiet && !opt_silent){
      for(int beeps=0;beeps<4;beeps++){
	  beep();
	  fflush(stdout);
	  Sleep(300);
      }
    }

    endwin();
    return 0;
}
