/*
 * Graphical user interface for AIMAGE
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology Corp.
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */



#include "afflib.h"			
#include "aimage.h"
#include "timer.h"
#include "imager.h"

bool show_help = false;
bool screen_cleared=false;
char *linebuf=0;

int rows = 0;
int cols = 0;

#ifndef CTRL
#define CTRL(x)  (x&037)
#endif

/* print val with commas.
 * buf must be at least 64 characters.
 */
void make_commas(uint64 val,char buf[64])
{
    char tmp[64];
    char t2[64];

    buf[0] = 0;
    if(val==0){
      strcpy(buf,"0");
    }

    while(val>0){
	int digits = val % 1000;		// get the residue
	val = val / 1000;		// and shift what's over

	if(val>0){			// we will still have more to do
	    sprintf(tmp,",%03d",digits);	// so pad it with zeros and a comma
	}
	else {
	    sprintf(tmp,"%d",digits);	// otherwise just get the value
	}
	strcpy(t2,buf);			// copy out the buffer
	strcpy(buf,tmp);		// copy in what we just did
	strcat(buf,t2);			// and put back what was there
    }
}

/* comma_printf():
 * Print a value with commas. If space!=0, allow this many characters
 * and right-justify
 */
void comma_printf(uint64 val,unsigned int space)
{
    char buf[64];
    make_commas(val,buf);
    while(strlen(buf) < space && space>0){
	putchar(' ');
	space--;
    }
    fputs(buf,stdout);			// send it to standard output
}



/****************************************************************
 *** Simson's termcap routines.
 ****************************************************************/

/* refresh():
 * Redraw the disk imaging status information...
 */
void new_line() 
{
    tputs(clr_eol,0,putchar);
    tputs(carriage_return,0,putchar);
    tputs(newline,0,putchar);
}

void bold(const char *str)
{
    if(str[0]){
	tputs(enter_bold_mode,1,putchar);
	fputs(str,stdout);
	tputs(exit_attribute_mode,0,putchar);
    }
}


void gotorc(int row,int col)
{
    char buf[256];
    char *cc = buf;
    tputs(tgoto(tgetstr("cm",&cc),col,row),0,putchar);
}


void rightflush(int row,const char *str)
{
    gotorc(row,cols-(strlen(str)+1));
    fputs(str,stdout);
}

void putchar_rev(char ch)
{
  tputs(enter_reverse_mode,0,putchar);
  putchar(ch);
  tputs(exit_attribute_mode,0,putchar);
}

void my_keyboard()
{
    switch(getch()){
    case ERR:
	return;
    case CTRL('l'):
	screen_cleared = false;		// need to clear the screen
	return;
    }
}

void my_refresh()
{
    char buf[1024];			// scratch

    my_keyboard();			// process keyboard commands

    /* Get the current screen size */
    const unsigned rows = tgetnum("li");
    const unsigned cols = tgetnum("co");

    const unsigned arrow_row   = 10;
    const unsigned current_row = 12;
    const unsigned reading_row = 15;
    const unsigned writing_row = 18;
    const unsigned space_row   = 21;
    const unsigned preview_row  = 6;
    const unsigned preview_rows = 4;


    static unsigned old_status_col = 0;

    /* This section draws the stuff that doesn't change very much.
     * It's faster not to redraw it all the time.
     */
    if(show_help){
	//display_help();
	show_help = false;	
	return;
    }

    if(screen_cleared == false){
	if(linebuf){
	    free(linebuf);
	    linebuf = (char *)malloc(cols+1);
	}

	tputs(clear_screen,0,putchar);
	screen_cleared = true;

	/* First time through, paint the stuff that doesn't change much */
    
	/* Source Device */
	int ncol = 16;		// where to put the first numeric column
	gotorc(1,1);    printf("Source device: ");
	gotorc(1,ncol); bold(infile);
	
	gotorc(2,   1);  printf("Model #:");
	gotorc(2,ncol);  bold(device_model);
	if(serial_number[0]){
	    gotorc(3,1);
	    printf("S/N:");
	    gotorc(3,ncol);
	    bold(serial_number);
	}
	if(firmware_revision[0]){
	    gotorc(4,1);
	    printf("firmware: ");
	    gotorc(4,ncol);
	    bold(firmware_revision);
	}
	
	/* Output */
	if(im->af){
	    gotorc(1,38); printf("AFF Output: "); bold(af_filename(im->af));
	}
	if(im->fout){
	    gotorc(0,40); printf("Raw Output: "); bold(im->fout_fname);
	}

	gotorc(3,35);  printf("    Disk Size:  ");
	uint64 disk_size = im->sector_size * im->total_sectors;
	if(disk_size>1000000000){
	    printf("%d GB",(int)(disk_size/1000000000));
	} else if(disk_size>1000000){
	    printf("%d MB",(int)(disk_size/1000000));
	} else if(disk_size>1024){
	    printf("%d KB",(int)(disk_size/1024));
	} else printf("?? ");
	
	printf(" (%d byte sectors)",im->sector_size);new_line();

	gotorc(4,35);  printf("Total sectors: ");
	if(im->total_sectors){
	  comma_printf(im->total_sectors,0);
	}
	else {
	    printf("????");
	}
	new_line();

	/* Print the bar graph, but only if we know how big the disk is */
	if(disk_size>0){
	    gotorc(arrow_row,0); putchar('[');
	    for(int i=0;i<cols-2;i++){
		putchar('-');
	    }
	    putchar(']');
	}
    }

    /* Stuff that changes a lot */

    gotorc(0,40-strlen(opt_title)/2);
    if(im->imaging){
      tputs(enter_blink_mode,0,putchar);
      tputs(enter_reverse_mode,0,putchar);
      printf("%s",opt_title);
      tputs(exit_attribute_mode,0,putchar);
    }
    else {
      printf("       ");
    }

	
    /* Display the time */
    char timebuf[64];
    time_t now = time(0);
    strcpy(timebuf,ctime(&now));
    timebuf[25] = '\000';
    rightflush(0,timebuf);
	
    gotorc(current_row,0);
    printf(" Currently reading sector: ");
    comma_printf(im->last_sector_read,11);
    printf(" (%3d sectors at once) ",im->last_sectors_read);

    gotorc(current_row+1,0); 
    printf("             Sectors read: ");
    comma_printf(im->total_sectors_read,11);

    double fraction_done = -1;
    if(im->total_sectors>0){		// can we figure this out?
	fraction_done = ((double)im->total_sectors_read / (double)im->total_sectors);
	printf(" (%5.2f%%) ", fraction_done * 100.0);
    }
    printf(" # blank: ");comma_printf(im->total_blank_sectors,0);
    if(im->total_sectors_read == im->total_blank_sectors){
	printf("  NO DATA YET!");
    }
    new_line();

    gotorc(reading_row,0);
    printf("       Time spent reading: %15s  ",
	   im->reading_timer.timer_text(buf));
    if(fraction_done>0){
	printf("   ETA: %s ",
	       total_time.eta_text(buf,fraction_done));
    }
    

    gotorc(reading_row+1,0);
    if(im->bad_sectors_read){
      printf("BAD SECTORS: ");comma_printf(im->bad_sectors_read,11);
      if(im->errors_in_a_row){
	printf(" (%d errors in a row)",im->errors_in_a_row);
      }
    }
    new_line();

    gotorc(reading_row+1,0);
    printf("         Total bytes read: ");
    comma_printf(im->total_bytes_read,15);
    
    gotorc(writing_row,0);
    printf("            bytes written: ");
    comma_printf(im->total_bytes_written,15);

    gotorc(writing_row+1,0);
    printf("   Time spent writing: %15s ", write_time.timer_text(buf));

    if(total_bytes_compressed>0 && total_bytes_uncompressed>0){
	gotorc(writing_row+2,0);
	printf(" Compressed bytes written: ");
	comma_printf(total_bytes_compressed,15);
	if(compressing){
	    printf("   ");
	    tputs(enter_blink_mode,0,putchar);
	    tputs(enter_reverse_mode,0,putchar);
	    printf(" >>> COMPRESSING <<< ");
	    tputs(exit_attribute_mode,0,putchar);
	}
	new_line();

	gotorc(writing_row+3,0);
	double overall_compression_ratio = 
	    (100.0 *(total_bytes_uncompressed-total_bytes_compressed))/ total_bytes_uncompressed;
	printf("Overall compression ratio: %6.2f%%  (0%% is none; 100%% is perfect)", overall_compression_ratio);
    }

    /* Redraw the arrow */
    if(im->total_sectors){
	unsigned new_status_col = (int)(((float)im->last_sector_read
				    /im->total_sectors) * (cols-2)) + 1;
	if(old_status_col && old_status_col != new_status_col){
	    /* Need to erase old status */
	    gotorc(arrow_row,old_status_col);
	    putchar_rev('=');
	}
	old_status_col = new_status_col;

	gotorc(arrow_row,new_status_col);
	if(im->last_direction==1){
	  putchar_rev('>');
	}
	else{
	  putchar_rev('<');
	}
    }

    /* Data preview... */
    if(opt_preview){
	unsigned int i;
	char b2[512];
	memcpy(b2,im->buf,512);
	for(i=0;i<sizeof(b2);i++){
	    if(!isprint(b2[i])) b2[i] = '.';
	}
	for(i=0;i<preview_rows;i++){
	    gotorc(preview_row+i,0);
	    fwrite(b2+i*80,1,79,stdout);
	}
    }

    int outfd = 0;
    if(im->af) outfd = fileno(im->af->aseg);
    if(im->fout) outfd = fileno(im->fout);
    if(outfd){
	gotorc(space_row,0);
	print_freespace(outfd);
	new_line();
    }

    gotorc(20,0);		// leave the cursor here for errors
    fflush(stdout);
}


